#ident	"@(#)stp.c	1.3"
/*
 * Copyright (c) 1993 Addison-Wesley Publishing Company.
 * All rights reserved.  This source code is provided
 * for its instructional value, and is not guaranteed
 * for any particular purpose.
 */
#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/errno.h>
#include <sys/cred.h>
#include <sys/byteorder.h>
#include <sys/cmn_err.h>
#include <sys/tihdr.h>
#include <sys/tiuser.h>
#include <sys/dlpi.h>
#include <sys/ddi.h>
#include <sys/stp.h>

#ifdef DEBUG
#define STATIC
#else
#define STATIC static
#endif

#define HALFSEC	500000	/* in microseconds */

#define HDRSIZE ((sizeof(hdr_t) < net.n_minsdu) ? \
	net.n_minsdu : sizeof(hdr_t))

int stpdevflag = 0;

/*
 * structure representing underlying network.
 */
enum netstate { DOWN, STATECHG, UP };

struct network {
	queue_t		*n_wrq;		/* upper write queue of network */
								/* (lower write queue of mux) */
	mblk_t		*n_mp;		/* ioctl message used */
								/* during link and unlink */
	enum netstate	n_state;	/* state of network */
	int		n_linkid;	/* link ID */
	uint_t		n_nextsend;	/* next minor to send a message */
	ulong_t		n_maxsdu;	/* max data unit size */
	ulong_t		n_minsdu;	/* min data unit size */
	char		n_saplen;	/* length of sap component of */
								/* network address */
	uchar_t		n_addrlen;	/* length of complete network */
								/* address */
	uchar_t		n_addr[NETADDRSZ+MAXNETSAPSZ];	/* network */
								/* (DLSAP) address */
};

STATIC struct network net;
STATIC dl_priority_t stpnopri = { 0, 0 };

/*
 * TPI state transition definitions.
 */
extern char ti_statetbl[TE_NOEVENTS][TS_NOSTATES];

#define NEXTSTATE1(CS, EV)  ti_statetbl[(EV)][(CS)]
#define NEXTSTATE2(CS, EV1, EV2) \
	ti_statetbl[(EV2)][ti_statetbl[(EV1)][(CS)]]
#define BADSTATE(CS, EV) \
	(ti_statetbl[(EV)][(CS)]==127)

/*
 * The following are defined in the driver
 * configuration file.
 */
extern int nstp;		/* number of endpoints */
extern stp_t stpendpt[];	/* array of endpoints */

typedef enum retval {
	DONE,		/* function completed its job */
	RETRY,		/* temporary failure; try again later */
	ERR		/* permanent error */
} retval_t;

STATIC int stpopen(queue_t *, dev_t *, int, int, cred_t *);
STATIC int stpclose(queue_t *, int, cred_t *);
STATIC int stpuwput(queue_t *, mblk_t *);
STATIC int stpuwsrv(queue_t *);
STATIC int stpursrv(queue_t *);
STATIC int stplwsrv(queue_t *);
STATIC int stplrput(queue_t *, mblk_t *);
STATIC int stplrsrv(queue_t *);
STATIC void stp_ioctl(queue_t *, mblk_t *);
STATIC void stp_wsched(stp_t *, int);
STATIC void stp_wcont(stp_t *);
STATIC retval_t stp_info(queue_t *, mblk_t *);
STATIC retval_t stp_bind(queue_t *, mblk_t *);
STATIC retval_t stp_unbind(queue_t *, mblk_t *);
STATIC retval_t stp_connect(queue_t *, mblk_t *);
STATIC retval_t stp_accept(queue_t *, mblk_t *);
STATIC retval_t stp_send(queue_t *, mblk_t *);
STATIC retval_t stp_discon(queue_t *, mblk_t *);
STATIC retval_t stp_optmgmt(queue_t *, mblk_t *);
STATIC retval_t stp_errorack(queue_t *, long, long, long);
STATIC retval_t stp_restart(stp_t *);
STATIC void stp_rsched(stp_t *, int);
STATIC void stp_rcont(stp_t *);
STATIC void stp_fatal(queue_t *, mblk_t *);
STATIC stp_t *stp_findport(ushort_t);
STATIC ushort_t stp_getuport(void);
STATIC void stp_xmit(stp_t *, mblk_t *, int);
STATIC void stp_retransmit(void);
STATIC void stp_disconall(uchar_t);
STATIC void stp_rcvinfo(mblk_t *);
STATIC retval_t stp_rcvdata(queue_t *, mblk_t *);
STATIC retval_t stp_rcvconnect(queue_t *, mblk_t *);
STATIC retval_t stp_rcvaccept(queue_t *, mblk_t *);
STATIC retval_t stp_rcvdiscon(queue_t *, mblk_t *);
STATIC void stp_buildokack(mblk_t *, ulong_t);
STATIC void stp_buildunitdata(mblk_t *, struct stpaddr *);
STATIC void stp_buildheader(stp_t *, mblk_t *, uchar_t,
	uchar_t, uchar_t, ushort_t, ushort_t, int);
STATIC void stp_buildaddr(struct stpaddr *, caddr_t,
	ushort_t);
STATIC void stp_discon1(stp_t *, uchar_t);
STATIC int stp_sendack(stp_t *, mblk_t *, uchar_t);
STATIC int stp_sameaddr(struct stpaddr *,
	struct stpaddr *);
STATIC int stp_samenetaddr(uchar_t *, uchar_t *);
STATIC int stp_qinuse(queue_t *);

STATIC struct module_info stp_minfo = {
	0x5354, "stp", 0, MAXTIDUSZ, 10*MAXTIDUSZ, 2*MAXTIDUSZ
};
STATIC struct qinit stp_urinit = {	/* upper read side */
	NULL, stpursrv, stpopen, stpclose, NULL, &stp_minfo, NULL
};
STATIC struct qinit stp_uwinit = {	/* upper write side */
	stpuwput, stpuwsrv, NULL, NULL, NULL, &stp_minfo, NULL
};
STATIC struct qinit stp_lrinit = {	/* lower read size */
	stplrput, stplrsrv, NULL, NULL, NULL, &stp_minfo, NULL
};
STATIC struct qinit stp_lwinit = {	/* lower write side */
	NULL, stplwsrv, NULL, NULL, NULL, &stp_minfo, NULL
};
struct streamtab stpinfo = {
	&stp_urinit, &stp_uwinit, &stp_lrinit, &stp_lwinit
};

void
stpinit()
{
	net.n_state = DOWN;
}

STATIC int
stpopen(queue_t *q, dev_t *devp, int flag, int sflag,
	cred_t *crp)
{
	stp_t *stp;
	int i;
	mblk_t *mp;
	struct stroptions *sp;

	if (sflag == MODOPEN)	/* configuration error */
		return(EINVAL);
	if (q->q_ptr != NULL)	/* not first open */
		return(0);
	if (sflag == CLONEOPEN) {
		for (i = 0; i < nstp; i++)
			if (stpendpt[i].s_rdq == NULL)
				break;
	} else {
		i = geteminor(*devp);
	}
	if (i >= nstp)
		return(ENXIO);
	mp = allocb(sizeof(struct stroptions), BPRI_HI);
	if (mp == NULL)
		return(EAGAIN);
	stp = &stpendpt[i];
	mp->b_datap->db_type = M_SETOPTS;
	mp->b_wptr += sizeof(struct stroptions);
	sp = (struct stroptions *)mp->b_rptr;
	sp->so_flags = SO_HIWAT|SO_LOWAT;
	sp->so_hiwat = stp_minfo.mi_hiwat;
	sp->so_lowat = stp_minfo.mi_lowat;
	putnext(q, mp);
	bzero((caddr_t)stp, sizeof(stp_t));
	stp->s_rdq = q;
	stp->s_tpstate = TS_UNBND;
	q->q_ptr = (caddr_t)stp;
	WR(q)->q_ptr = (caddr_t)stp;
	if (sflag == CLONEOPEN)
		*devp = makedevice(getemajor(*devp), i);
	return(0);
}

STATIC int
stpclose(queue_t *q, int flag, cred_t *crp)
{
	stp_t *stp;
	int i;

	stp = (stp_t *)q->q_ptr;
	if (stp->s_rxmp != NULL) {
		freemsg(stp->s_rxmp);
		stp->s_rxmp = NULL;
	}
	if ((stp->s_tpstate == TS_DATA_XFER) &&
	  (net.n_state == UP)) {
		stp_discon(WR(q), NULL);
		if (stp->s_rxmp != NULL) {
			freemsg(stp->s_rxmp);
			stp->s_rxmp = NULL;
		}
	}
	if (stp->s_rcvmp != NULL) {
		freemsg(stp->s_rcvmp);
		stp->s_rcvmp = NULL;
	}
	if (stp->s_flag & RRECBCALL)
		unbufcall(stp->s_rrecid);
	else if (stp->s_flag & RRECTOUT)
		untimeout(stp->s_rrecid);
	if (stp->s_flag & WRECBCALL)
		unbufcall(stp->s_wrecid);
	else if (stp->s_flag & WRECTOUT)
		untimeout(stp->s_wrecid);
	for (i = 0; i < MAXCREQ; i++) {
		if (stp->s_cmp[i] != NULL) {
			freeb(stp->s_cmp[i]);
			stp->s_cmp[i] = NULL;
		}
	}
	q->q_ptr = NULL;
	WR(q)->q_ptr = NULL;
	stp->s_rdq = NULL;
	return(0);
}

STATIC int
stpuwput(queue_t *q, mblk_t *mp)
{
	mblk_t *bp, *nbp;
	union T_primitives *tp;

	switch (mp->b_datap->db_type) {
	case M_PROTO:
	case M_PCPROTO:
		tp = (union T_primitives *)mp->b_rptr;
		if (tp->type == T_DISCON_REQ) {
			bp = q->q_first;
			while (bp != NULL) {
				nbp = bp->b_next;
				if (bp->b_datap->db_type == M_DATA) {
					rmvq(q, bp);
					freemsg(bp);
					bp = nbp;
					continue;
				}
				tp = (union T_primitives *)bp->b_rptr;
				switch (tp->type) {
				case T_DATA_REQ:
				case T_CONN_RES:
					rmvq(q, bp);
					freemsg(bp);
					break;

				case T_CONN_REQ:
					rmvq(q, bp);
					freemsg(bp);
					freemsg(mp);
					mp = NULL;
					break;
				}
				bp = nbp;
			}
		}
		if (mp == NULL)
			break;
		/* FALLTHRU */

	case M_DATA:
		putq(q, mp);
		break;

	case M_IOCTL:
		stp_ioctl(q, mp);
		break;

	case M_FLUSH:
		if (*mp->b_rptr & FLUSHW) {
			flushq(q, FLUSHALL);
			*mp->b_rptr &= ~FLUSHW;
		}
		if (*mp->b_rptr & FLUSHR) {
			flushq(RD(q), FLUSHALL);
			qreply(q, mp);
		} else {
			freemsg(mp);
		}
		break;

	default:
		freemsg(mp);
	}
	return(0);
}

STATIC void
stp_ioctl(queue_t *q, mblk_t *mp)
{
	struct iocblk *iocp;
	struct linkblk *lp;
	dl_info_req_t *dp;
	mblk_t *bp;

	iocp = (struct iocblk *)mp->b_rptr;
	switch (iocp->ioc_cmd) {
	case I_LINK:
	case I_PLINK:
		bp = mp->b_cont;
		if ((bp == NULL) || ((bp->b_wptr - bp->b_rptr) !=
		  sizeof(struct linkblk)))
			goto nak;
		lp = (struct linkblk *)bp->b_rptr;
		if (net.n_state != DOWN) {
			iocp->ioc_error = EBUSY;
			goto nak;
		}
		bp = allocb(DL_INFO_REQ_SIZE, BPRI_HI);
		if (bp == NULL) {
			iocp->ioc_error = EAGAIN;
			goto nak;
		}
		if (iocp->ioc_cmd == I_PLINK)
			lp->l_qtop = q;
		net.n_mp = mp;
		net.n_wrq = lp->l_qbot;
		net.n_linkid = lp->l_index;
		net.n_state = STATECHG;
		bp->b_datap->db_type = M_PCPROTO;
		bp->b_wptr += DL_INFO_REQ_SIZE;
		dp = (dl_info_req_t *)bp->b_rptr;
		dp->dl_primitive = DL_INFO_REQ;
		putnext(net.n_wrq, bp);
		break;

	case I_UNLINK:
	case I_PUNLINK:
		bp = mp->b_cont;
		if ((bp == NULL) || ((bp->b_wptr - bp->b_rptr) !=
		  sizeof(struct linkblk)))
			goto nak;
		lp = (struct linkblk *)bp->b_rptr;
		if (net.n_state != UP) {
			iocp->ioc_error = ENONET;
			goto nak;
		}
		if (lp->l_index != net.n_linkid)
			goto nak;
		net.n_wrq = NULL;
		net.n_state = DOWN;
		mp->b_cont = NULL;
		freeb(bp);
		stp_disconall(DIS_NOLINK);
		mp->b_datap->db_type = M_IOCACK;
		iocp->ioc_count = 0;
		qreply(q, mp);
		break;

	default:
nak:
		mp->b_datap->db_type = M_IOCNAK;
		qreply(q, mp);
	}
}

STATIC int
stpuwsrv(queue_t *q)
{
	mblk_t *mp;
	stp_t *stp;
	union T_primitives *tp;
	retval_t ret;
	int again;

	stp = (stp_t *)q->q_ptr;
	if (net.n_state != UP) {
		flushq(q, FLUSHALL);
		return(0);
	}
	if (stp->s_rxmp != NULL) {
		if ((stp->s_totxmitsz != stp->s_msgxmitsz) &&
		  (stp->s_rcvack == (stp->s_seqno - 1)))
			stp_xmit(stp, NULL, 0);
		return(0);
	}
	if (stp->s_flag & DORESTART) {
		stp_restart(stp);
		return(0);
	}
	while ((mp = getq(q)) != NULL) {
		again = 0;
		if (mp->b_datap->db_type == M_DATA) {
			ret = stp_send(q, mp);
		} else {
			tp = (union T_primitives *)mp->b_rptr;
			switch (tp->type) {
			case T_DATA_REQ:
				ret = stp_send(q, mp);
				break;

			case T_INFO_REQ:
				again = 1;
				ret = stp_info(q, mp);
				break;

			case T_BIND_REQ:
				again = 1;
				ret = stp_bind(q, mp);
				break;

			case T_UNBIND_REQ:
				again = 1;
				ret = stp_unbind(q, mp);
				break;

			case T_CONN_REQ:
				ret = stp_connect(q, mp);
				break;

			case T_CONN_RES:
				again = 1;
				ret = stp_accept(q, mp);
				break;

			case T_DISCON_REQ:
				ret = stp_discon(q, mp);
				break;

			case T_OPTMGMT_REQ:
				again = 1;
				ret = stp_optmgmt(q, mp);
				break;

			default:
				stp_fatal(q, mp);
				return(0);
			}
		} /* else */
		if (ret == RETRY) {
			putbq(q, mp);
			return(0);
		} else if ((ret == DONE) && !again) {
			return(0);
		}
	} /* while */
	return(0);
}

STATIC retval_t
stp_info(queue_t *q, mblk_t *mp)
{
	mblk_t *bp;
	stp_t *stp;
	struct T_info_ack *ackp;

	freemsg(mp);
	stp = (stp_t *)q->q_ptr;
	bp = allocb(sizeof(struct T_info_ack), BPRI_HI);
	if (bp == NULL) {
		cmn_err(CE_WARN, "TCP: can't respond to T_INFO_REQ");
		return(DONE);
	}
	bp->b_datap->db_type = M_PCPROTO;
	bp->b_wptr += sizeof(struct T_info_ack);
	ackp = (struct T_info_ack *)bp->b_rptr;
	ackp->PRIM_type = T_INFO_ACK;
	ackp->TSDU_size = -1;	/* infinite */
	ackp->ETSDU_size = -2;	/* unsupported */
	ackp->CDATA_size = -2;	/* no data on connect */
	ackp->DDATA_size = -2;	/* no data on disconnect */
	ackp->ADDR_size = STPADDRSZ;
	ackp->OPT_size = -2;	/* no user-settable options */
	ackp->TIDU_size = MAXTIDUSZ;
	ackp->SERV_type = T_COTS;
	ackp->CURRENT_state = stp->s_tpstate;
	ackp->PROVIDER_flag = TP_SNDZERO;
	qreply(q, mp);
	return(DONE);
}

STATIC void
stp_wsched(stp_t *stp, int size)
{
	ulong_t id;
	int s;

	if (stp->s_flag & (WRECBCALL|WRECTOUT))
		return;
	id = bufcall(size, BPRI_HI, stp_wcont, stp);
	s = splstr();
	if (id == 0) {
		id = timeout(stp_wcont, (caddr_t)stp,
		  drv_usectohz(HALFSEC));
		stp->s_flag |= WRECTOUT;
	} else {
		stp->s_flag |= WRECBCALL;
	}
	splx(s);
	stp->s_wrecid = id;
}

STATIC void
stp_wcont(stp_t *stp)
{
	if (stp->s_rdq != NULL) {
		qenable(WR(stp->s_rdq));
		stp->s_flag &= ~(WRECBCALL|WRECTOUT);
	}
}

STATIC retval_t
stp_bind(queue_t *q, mblk_t *mp)
{
	stp_t *stp;
	mblk_t *bp;
	ushort_t port;
	struct T_bind_req *reqp;
	struct T_bind_ack *ackp;
	struct stpaddr *ap;
	long terr;
	int i;

	stp = (stp_t *)q->q_ptr;
	bp = NULL;
	if (BADSTATE(stp->s_tpstate, TE_BIND_REQ)) {
		terr = TOUTSTATE;
		goto err;
	}
	reqp = (struct T_bind_req *)mp->b_rptr;
	if (((reqp->ADDR_offset != 0) &&
	  ((reqp->ADDR_offset < sizeof(struct T_bind_req)) ||
	  (reqp->ADDR_length != STPADDRSZ))) ||
	  ((mp->b_wptr - mp->b_rptr) < (reqp->ADDR_offset +
	  reqp->ADDR_length))) {
		terr = TBADADDR;
		goto err;
	}
	bp = allocb(sizeof(struct T_bind_ack)+STPADDRSZ,
	  BPRI_HI);
	if (bp == NULL) {
		stp_wsched(stp, sizeof(struct T_bind_ack)+STPADDRSZ);
		return(RETRY);
	}
	if (reqp->CONIND_number > MAXCREQ)
		reqp->CONIND_number = MAXCREQ;
	if (reqp->ADDR_length == 0) {
		if ((port = stp_getuport()) == 0) {
			terr = TNOADDR;	/* all ports in use */
			goto err;
		}
	} else {
		ap = (struct stpaddr *)(mp->b_rptr+reqp->ADDR_offset);
		port = ntohs(ap->a_port);
		if (port == 0) {
			terr = TBADADDR;
			goto err;
		}
		for (i = 0; i < NETADDRSZ; i++)
			if (ap->a_net[i] != net.n_addr[i])
				break;
		if (i < NETADDRSZ) {
			terr = TBADADDR;
			goto err;
		}
		if (stp_findport(port) != NULL) {
			terr = TNOADDR;
			goto err;
		}
	}
	bp->b_datap->db_type = M_PCPROTO;
	ackp = (struct T_bind_ack *)bp->b_wptr;
	ackp->PRIM_type = T_BIND_ACK;
	ackp->ADDR_length = STPADDRSZ;
	ackp->ADDR_offset = sizeof(struct T_bind_ack);
	ackp->CONIND_number = reqp->CONIND_number;
	bp->b_wptr += sizeof(struct T_bind_ack);
	*(ushort_t *)bp->b_wptr = htons(port);
	bp->b_wptr += 2;
	bcopy((caddr_t)net.n_addr, (caddr_t)bp->b_wptr, NETADDRSZ);
	bp->b_wptr += NETADDRSZ;
	stp->s_lport = port;
	stp->s_maxqlen = reqp->CONIND_number;
	stp->s_tpstate = NEXTSTATE2(stp->s_tpstate, TE_BIND_REQ,
	  TE_BIND_ACK);
	freemsg(mp);
	qreply(q, bp);
	return(DONE);
err:
	if (bp != NULL)
		freeb(bp);
	if (stp_errorack(q, T_BIND_REQ, terr, 0) == RETRY)
		return(RETRY);
	freemsg(mp);
	return(ERR);
}

STATIC retval_t
stp_unbind(queue_t *q, mblk_t *mp)
{
	stp_t *stp;
	mblk_t *bp;
	int s;

	stp = (stp_t *)q->q_ptr;
	if (BADSTATE(stp->s_tpstate, TE_UNBIND_REQ)) {
		if (stp_errorack(q, T_UNBIND_REQ, TOUTSTATE, 0) ==
		  RETRY)
			return(RETRY);
		freemsg(mp);
		return(ERR);
	}
	bp = allocb(sizeof(struct T_ok_ack), BPRI_HI);
	if (bp == NULL) {
		stp_wsched(stp, sizeof(struct T_ok_ack));
		return(RETRY);
	}
	if (putctl1(RD(q)->q_next, M_FLUSH, FLUSHRW) == 0) {
		freeb(bp);
		if (!(stp->s_flag&(WRECBCALL|WRECTOUT))) {
			stp->s_wrecid = timeout(stp_wcont, (caddr_t)stp,
			  drv_usectohz(HALFSEC));
			s = splstr();
			stp->s_flag |= WRECTOUT;
			splx(s);
		}
		return(RETRY);
	}
	freemsg(mp);
	stp_buildokack(bp, T_UNBIND_REQ);
	stp->s_lport = 0;
	stp->s_maxqlen = 0;
	s = splstr();
	if (stp->s_rxmp != NULL) {
		freemsg(stp->s_rxmp);
		stp->s_rxmp = NULL;
	}
	splx(s);
	if (stp->s_rcvmp != NULL) {
		freemsg(stp->s_rcvmp);
		stp->s_rcvmp = NULL;
	}
	stp->s_tpstate = NEXTSTATE2(stp->s_tpstate, TE_UNBIND_REQ,
	  TE_OK_ACK1);
	qreply(q, bp);
	return(DONE);
}

STATIC void
stp_buildokack(mblk_t *bp, ulong_t prim)
{
	struct T_ok_ack *ackp;

	bp->b_datap->db_type = M_PCPROTO;
	ackp = (struct T_ok_ack *)bp->b_wptr;
	ackp->PRIM_type = T_OK_ACK;
	ackp->CORRECT_prim = prim;
	bp->b_wptr += sizeof(struct T_ok_ack);
}

STATIC retval_t
stp_connect(queue_t *q, mblk_t *mp)
{
	stp_t *stp;
	mblk_t *bp, *hbp, *dbp;
	struct T_conn_req *reqp;
	struct stpaddr *ap;
	int size;
	long terr, uerr;

	stp = (stp_t *)q->q_ptr;
	uerr = 0;
	bp = hbp = dbp = NULL;
	if (BADSTATE(stp->s_tpstate, TE_CONN_REQ)) {
		terr = TOUTSTATE;
		goto err;
	}
	reqp = (struct T_conn_req *)mp->b_rptr;
	if ((reqp->DEST_offset < sizeof(struct T_conn_req)) ||
	  (reqp->DEST_length != STPADDRSZ) ||
	  ((mp->b_wptr - mp->b_rptr) < (reqp->DEST_offset +
	  reqp->DEST_length))) {
		terr = TBADADDR;
		goto err;
	}
	if (reqp->OPT_length != 0) {
		terr = TBADOPT;
		goto err;
	}
	if (mp->b_cont != NULL) {
		terr = TBADDATA;
		goto err;
	}
	if (!canput(net.n_wrq->q_next))
		return(RETRY);
	bp = allocb(sizeof(struct T_ok_ack), BPRI_HI);
	if (bp == NULL) {
		stp_wsched(stp, sizeof(struct T_ok_ack));
		return(RETRY);
	}
	size = HDRSIZE;
	hbp = allocb(size, BPRI_HI);
	dbp = allocb(DL_UNITDATA_REQ_SIZE+net.n_addrlen,
	  BPRI_HI);
	if ((hbp == NULL) || (dbp == NULL)) {
		terr = TSYSERR;
		uerr = EAGAIN;
		goto err;
	}
	ap = (struct stpaddr *)(mp->b_rptr + reqp->DEST_offset);
	stp_buildunitdata(dbp, ap);
	stp_buildheader(stp, hbp, CONNECT, EOM, 0, sizeof(hdr_t),
	  ap->a_port, size);
	dbp->b_cont = hbp;
	stp_xmit(stp, dbp, 0);
	stp_buildokack(bp, T_CONN_REQ);
	stp->s_faddr = *ap;	/* structure assignment */
	freemsg(mp);
	stp->s_tpstate = NEXTSTATE2(stp->s_tpstate, TE_CONN_REQ,
	  TE_OK_ACK1);
	qreply(q, bp);
	return(DONE);
err:
	if (bp != NULL)
		freeb(bp);
	if (hbp != NULL)
		freeb(hbp);
	if (stp_errorack(q, T_CONN_REQ, terr, uerr) == RETRY)
		return(RETRY);
	freemsg(mp);
	return(ERR);
}

STATIC void
stp_buildunitdata(mblk_t *bp, struct stpaddr *ap)
{
	dl_unitdata_req_t *udp;

	bp->b_datap->db_type = M_PROTO;
	bp->b_wptr += DL_UNITDATA_REQ_SIZE;
	udp = (dl_unitdata_req_t *)bp->b_rptr;
	udp->dl_primitive = DL_UNITDATA_REQ;
	udp->dl_dest_addr_length = net.n_addrlen;
	udp->dl_dest_addr_offset = DL_UNITDATA_REQ_SIZE;
	udp->dl_priority = stpnopri;
	if (net.n_saplen > 0) {	/* sap followed by phys */
		bcopy((caddr_t)net.n_addr, (caddr_t)bp->b_wptr,
		  net.n_saplen);
		bp->b_wptr += net.n_saplen;
		bcopy((caddr_t)ap->a_net, (caddr_t)bp->b_wptr,
		  NETADDRSZ);
		bp->b_wptr += NETADDRSZ;
	} else if (net.n_saplen < 0) {	/* phys followed by sap */
		bcopy((caddr_t)ap->a_net, (caddr_t)bp->b_wptr,
		  NETADDRSZ);
		bp->b_wptr += NETADDRSZ;
		bcopy((caddr_t)&net.n_addr[net.n_addrlen+net.n_saplen],
		  (caddr_t)bp->b_wptr, -net.n_saplen);
		bp->b_wptr -= net.n_saplen;
	} else {
		bcopy((caddr_t)ap->a_net, (caddr_t)bp->b_wptr,
		  NETADDRSZ);
		bp->b_wptr += NETADDRSZ;
	}
}

STATIC void
stp_buildheader(stp_t *stp, mblk_t *bp, uchar_t type,
	uchar_t flags, uchar_t reason, ushort_t totsz,
	ushort_t dport, int mpsize)
{
	hdr_t *hp;

	hp = (hdr_t *)bp->b_wptr;
	hp->h_version = STPVERS;
	hp->h_type = type;
	hp->h_flags = flags;
	hp->h_reason = reason;
	hp->h_sport = htons(stp->s_lport);
	hp->h_dport = dport;	/* already in network byte order */
	hp->h_size = htons(totsz);
	hp->h_pad = 0;
	hp->h_seqno = 0;	/* filled in by stp_xmit */
	bp->b_wptr += mpsize;
}

STATIC retval_t
stp_accept(queue_t *q, mblk_t *mp)
{
	stp_t *stp, *tstp;
	mblk_t *bp, *hbp, *dbp;
	struct T_conn_res *resp;
	struct T_conn_ind *indp;
	struct stpaddr *ap;
	int size, i, cidx;
	long terr, uerr;

	stp = (stp_t *)q->q_ptr;
	uerr = 0;
	bp = hbp = dbp = NULL;
	if (BADSTATE(stp->s_tpstate, TE_CONN_RES)) {
		terr = TOUTSTATE;
		goto err;
	}
	resp = (struct T_conn_res *)mp->b_rptr;
	if ((stp->s_qlen > 1) &&
	  (resp->QUEUE_ptr == stp->s_rdq)) {
		terr = TBADF;
		goto err;
	}
	if (resp->OPT_length != 0) {
		terr = TBADOPT;
		goto err;
	}
	if (mp->b_cont != NULL) {
		terr = TBADDATA;
		goto err;
	}
	if (!canput(net.n_wrq->q_next))
		return(RETRY);
	for (cidx = 0; cidx < MAXCREQ; cidx++) {
		if (stp->s_cmp[cidx] != NULL) {
			indp = (struct T_conn_ind *)
			  stp->s_cmp[cidx]->b_rptr;
			if (resp->SEQ_number == indp->SEQ_number)
				break;
		}
	}
	if (cidx >= MAXCREQ) {
		terr = TBADSEQ;
		goto err;
	}
	ap = (struct stpaddr *)(stp->s_cmp[cidx]->b_rptr +
	  indp->SRC_offset);
	bp = allocb(sizeof(struct T_ok_ack), BPRI_HI);
	if (bp == NULL) {
		stp_wsched(stp, sizeof(struct T_ok_ack));
		return(RETRY);
	}
	size = HDRSIZE;
	hbp = allocb(size, BPRI_HI);
	dbp = allocb(DL_UNITDATA_REQ_SIZE+net.n_addrlen,
	  BPRI_HI);
	if ((hbp == NULL) || (dbp == NULL)) {
		terr = TSYSERR;
		uerr = EAGAIN;
		goto err;
	}
	for (tstp = stpendpt, i = 0; i < nstp; tstp++, i++)
		if (tstp->s_rdq == resp->QUEUE_ptr)
			break;
	if (i >= nstp) {
		terr = TBADF;
		goto err;
	}
	if (tstp != stp) {
		if (tstp->s_tpstate != TS_IDLE) {
			terr = TBADF;
			goto err;
		}
		tstp->s_tpstate = NEXTSTATE1(tstp->s_tpstate,
		  TE_PASS_CONN);
	}
	tstp->s_faddr = *ap;	/* structure assignment */
	freeb(stp->s_cmp[cidx]);
	stp->s_cmp[cidx] = NULL;
	stp->s_qlen--;
	stp_buildunitdata(dbp, &tstp->s_faddr);
	stp_buildheader(tstp, hbp, ACCEPT, EOM, 0, sizeof(hdr_t),
	  tstp->s_faddr.a_port, size);
	dbp->b_cont = hbp;
	stp_xmit(tstp, dbp, 0);
	stp_buildokack(bp, T_CONN_RES);
	freemsg(mp);
	if (stp->s_qlen == 0) {
		if (stp == tstp)
			stp->s_tpstate = NEXTSTATE2(stp->s_tpstate,
			  TE_CONN_RES, TE_OK_ACK2);
		else
			stp->s_tpstate = NEXTSTATE2(stp->s_tpstate,
			  TE_CONN_RES, TE_OK_ACK3);
	} else {
		stp->s_tpstate = NEXTSTATE2(stp->s_tpstate,
		  TE_CONN_RES, TE_OK_ACK4);
	}
	qreply(q, bp);
	return(DONE);
err:
	if (bp != NULL)
		freeb(bp);
	if (hbp != NULL)
		freeb(hbp);
	if (dbp != NULL)
		freeb(dbp);
	if (stp_errorack(q, T_CONN_RES, terr, uerr) == RETRY)
		return(RETRY);
	freemsg(mp);
	return(ERR);
}

STATIC int
stp_sameaddr(struct stpaddr *a1, struct stpaddr *a2)
{
	int i;

	for (i = 0; i < NETADDRSZ; i++)
		if (a1->a_net[i] != a2->a_net[i])
			return(0);
	if (a1->a_port != a2->a_port)
		return(0);
	return(1);
}

STATIC retval_t
stp_discon(queue_t *q, mblk_t *mp)
{
	stp_t *stp;
	mblk_t *bp, *hbp, *dbp;
	struct T_discon_req *reqp;
	struct T_conn_ind *indp;
	struct stpaddr addr;
	int size, i, s;
	long terr, uerr;

	stp = (stp_t *)q->q_ptr;
	uerr = 0;
	bp = hbp = dbp = NULL;
	if (BADSTATE(stp->s_tpstate, TE_DISCON_REQ)) {
		terr = TOUTSTATE;
		goto err;
	}
	if ((mp != NULL) && (mp->b_cont != NULL)) {
		terr = TBADDATA;
		goto err;
	}
	if (!canput(net.n_wrq->q_next))
		return(RETRY);
	bp = allocb(sizeof(struct T_ok_ack), BPRI_HI);
	if (bp == NULL) {
		stp_wsched(stp, sizeof(struct T_ok_ack));
		return(RETRY);
	}
	size = HDRSIZE;
	hbp = allocb(size, BPRI_HI);
	dbp = allocb(DL_UNITDATA_REQ_SIZE+net.n_addrlen,
	  BPRI_HI);
	if ((hbp == NULL) || (dbp == NULL)) {
		terr = TSYSERR;
		uerr = EAGAIN;
		goto err;
	}
	if (mp != NULL)
		reqp = (struct T_discon_req *)mp->b_rptr;
	if (stp->s_tpstate == TS_WRES_CIND) {
		for (i = 0; i < MAXCREQ; i++) {
			if (stp->s_cmp[i] != NULL) {
				indp = (struct T_conn_ind *)
				  stp->s_cmp[i]->b_rptr;
				if (reqp->SEQ_number == indp->SEQ_number)
					break;
			}
		}
		if (i >= MAXCREQ) {
			terr = TBADSEQ;
			goto err;
		}
		bcopy((caddr_t)stp->s_cmp[i]->b_rptr+indp->SRC_offset,
		  (caddr_t)&addr, STPADDRSZ);
		freeb(stp->s_cmp[i]);
		stp->s_cmp[i] = NULL;
		stp->s_qlen--;
		if (stp->s_qlen == 0)
			stp->s_tpstate = NEXTSTATE2(stp->s_tpstate,
			  TE_DISCON_REQ, TE_OK_ACK2);
		else
			stp->s_tpstate = NEXTSTATE2(stp->s_tpstate,
			  TE_DISCON_REQ, TE_OK_ACK4);
	} else {
		if (stp->s_tpstate == TS_DATA_XFER) {
			if (putctl1(RD(q)->q_next, M_FLUSH, FLUSHRW) ==
			  0) {
				freeb(bp);
				freeb(hbp);
				freeb(dbp);
				if (!(stp->s_flag&(WRECBCALL|WRECTOUT))) {
					stp->s_wrecid = timeout(stp_wcont,
					  (caddr_t)stp, drv_usectohz(HALFSEC));
					s = splstr();
					stp->s_flag |= WRECTOUT;
					splx(s);
				}
				return(RETRY);
			}
			addr = stp->s_faddr;	/* structure assignment */
		}
		stp->s_tpstate = NEXTSTATE2(stp->s_tpstate,
		  TE_DISCON_REQ, TE_OK_ACK1);
	}
	stp_buildunitdata(dbp, &addr);
	stp_buildheader(stp, hbp, DISCON, EOM, DIS_USER,
	  sizeof(hdr_t), addr.a_port, size);
	dbp->b_cont = hbp;
	stp_xmit(stp, dbp, 0);
	stp_buildokack(bp, T_DISCON_REQ);
	if (mp != NULL)
		freemsg(mp);
	qreply(q, bp);
	bzero((caddr_t)&stp->s_faddr, STPADDRSZ);
	return(DONE);
err:
	if (bp != NULL)
		freeb(bp);
	if (hbp != NULL)
		freeb(hbp);
	if (dbp != NULL)
		freeb(dbp);
	if (stp_errorack(q, T_DISCON_REQ, terr, uerr) == RETRY)
		return(RETRY);
	if (mp != NULL)
		freemsg(mp);
	return(ERR);
}

STATIC retval_t
stp_send(queue_t *q, mblk_t *mp)
{
	stp_t *stp;
	mblk_t *bp, *hbp, *dbp, *pbp;
	int size, psize, s;
	struct T_data_req *reqp;
	uchar_t flags = 0;

	stp = (stp_t *)q->q_ptr;
	s = splstr();
	if (stp->s_flag & NOXMIT) {
		splx(s);
		return(RETRY);
	}
	splx(s);
	if (stp->s_tpstate == TS_IDLE) {
		freemsg(mp);
		return(ERR);
	}
	if (BADSTATE(stp->s_tpstate, TE_DATA_REQ)) {
		stp_fatal(q, mp);
		return(ERR);
	}
	if (!canput(net.n_wrq->q_next))
		return(RETRY);
	size = msgdsize(mp) + sizeof(hdr_t);
	psize = 0;
	if (size < net.n_minsdu) {
		psize = net.n_minsdu - size;
		size = net.n_minsdu;
	}
	hbp = allocb(sizeof(hdr_t), BPRI_MED);
	if (hbp == NULL) {
		stp_wsched(stp, sizeof(hdr_t));
		return(RETRY);
	}
	dbp = allocb(DL_UNITDATA_REQ_SIZE+net.n_addrlen,
	  BPRI_MED);
	if (dbp == NULL)
		goto allocfail;
	if (psize != 0) {
		pbp = allocb(psize, BPRI_MED);
		if (pbp == NULL)
			goto allocfail;
		pbp->b_wptr += psize;
		linkb(mp, pbp);
	}
	if (mp->b_datap->db_type == M_PROTO) {
		reqp = (struct T_data_req *)mp->b_rptr;
		if (reqp->MORE_flag != 0)
			flags |= MORE;
		bp = mp;
		mp = mp->b_cont;
		freeb(bp);
	}
	stp_buildunitdata(dbp, &stp->s_faddr);
	stp_buildheader(stp, hbp, DATA, flags, 0, size - psize,
	  stp->s_faddr.a_port, sizeof(hdr_t));
	hbp->b_cont = mp;
	dbp->b_cont = hbp;
	stp_xmit(stp, dbp, 0);
	return(DONE);

allocfail:
	freeb(hbp);
	if (dbp != NULL)
		freeb(dbp);
	if (!(stp->s_flag&(WRECBCALL|WRECTOUT))) {
		stp->s_wrecid = timeout(stp_wcont, (caddr_t)stp,
		  drv_usectohz(HALFSEC));
		s = splstr();
		stp->s_flag |= WRECTOUT;
		splx(s);
	}
	return(RETRY);
}

STATIC retval_t
stp_restart(stp_t *stp)
{
	mblk_t *hbp, *dbp;
	int size;

	if (net.n_state != UP)
		return(ERR);
	if (!(stp->s_flag & DORESTART))
		return(ERR);
	if (!canput(net.n_wrq->q_next))
		return(RETRY);
	if (stp->s_rxmp != NULL)
		return(RETRY);
	size = HDRSIZE;
	hbp = allocb(size, BPRI_MED);
	if (hbp == NULL)
		return(RETRY);
	dbp = allocb(DL_UNITDATA_REQ_SIZE+net.n_addrlen,
	  BPRI_MED);
	if (dbp == NULL) {
		freeb(hbp);
		return(RETRY);
	}
	stp_buildunitdata(dbp, &stp->s_faddr);
	stp_buildheader(stp, hbp, RESTART, EOM, 0,
	  sizeof(hdr_t), stp->s_faddr.a_port, size);
	dbp->b_cont = hbp;
	stp_xmit(stp, dbp, 0);
	stp->s_flag &= ~DORESTART;
	return(DONE);
}

STATIC retval_t
stp_optmgmt(queue_t *q, mblk_t *mp)
{
	stp_t *stp;
	struct T_optmgmt_req *reqp;	/* we know these */
	struct T_optmgmt_ack *ackp;	/* are the same size */
	long flags, terr;

	stp = (stp_t *)q->q_ptr;
	if (BADSTATE(stp->s_tpstate, TE_OPTMGMT_REQ)) {
		terr = TOUTSTATE;
		goto err;
	}
	reqp = (struct T_optmgmt_req *)mp->b_rptr;
	if (reqp->OPT_length != 0) {
		terr = TBADOPT;
		goto err;
	}
	flags = reqp->MGMT_flags;
	terr = 0;
	switch (flags) {
	case T_NEGOTIATE:
	case T_CHECK:
		break;

	case T_DEFAULT:
		flags |= T_SUCCESS;
		break;

	default:
		terr = TBADFLAG;
	}
	if (terr != 0)
		goto err;
	mp->b_datap->db_type = M_PCPROTO;
	ackp = (struct T_optmgmt_ack *)mp->b_rptr;
	ackp->PRIM_type = T_OPTMGMT_ACK;
	ackp->OPT_length = 0;
	ackp->OPT_offset = 0;
	ackp->MGMT_flags = flags;
	qreply(q, mp);
	return(DONE);
err:
	if (stp_errorack(q, T_OPTMGMT_REQ, terr, 0) == RETRY)
		return(RETRY);
	freemsg(mp);
	return(ERR);
}

STATIC retval_t
stp_errorack(queue_t *q, long prim, long terr, long uerr)
{
	mblk_t *bp;
	stp_t *stp;
	struct T_error_ack *ackp;

	stp = (stp_t *)q->q_ptr;
	bp = allocb(sizeof(struct T_error_ack), BPRI_HI);
	if (bp == NULL) {
		stp_wsched(stp, sizeof(struct T_error_ack));
		return(RETRY);
	}
	bp->b_datap->db_type = M_PCPROTO;
	bp->b_wptr += sizeof(struct T_error_ack);
	ackp = (struct T_error_ack *)bp->b_rptr;
	ackp->PRIM_type = T_ERROR_ACK;
	ackp->ERROR_prim = prim;
	ackp->TLI_error = terr;
	ackp->UNIX_error = uerr;
	qreply(q, bp);
	return(DONE);
}

STATIC void
stp_fatal(queue_t *q, mblk_t *mp)
{
	if (mp->b_cont != NULL) {
		freemsg(mp->b_cont);
		mp->b_cont = NULL;
	}
	mp->b_datap->db_type = M_ERROR;
	mp->b_rptr = mp->b_datap->db_base;
	*mp->b_rptr = EPROTO;
	mp->b_wptr = mp->b_rptr + 1;
	qreply(q, mp);
}

STATIC void
stp_xmit(stp_t *stp, mblk_t *mp, int retransmit)
{
	mblk_t *dupmp, *bp, *pbp;
	hdr_t *hp;
	int s, size, skip, maxsz;

	s = splstr();
	if (mp != NULL) {
		stp->s_rxmp = mp;
		stp->s_nretr = 0;
		stp->s_totxmitsz = 0;
		hp = (hdr_t *)mp->b_cont->b_rptr;
		stp->s_msgxmitsz = ntohs(hp->h_size);
	} else {
		if (!retransmit)
			stp->s_nretr = 0;
		if ((stp->s_rxmp == NULL) ||
		  !canput(net.n_wrq->q_next)) {
			splx(s);
			return;
		}
		hp = (hdr_t *)stp->s_rxmp->b_cont->b_rptr;
	}
	dupmp = dupmsg(stp->s_rxmp);
	splx(s);
	if (dupmp == NULL)
		return;
	if (stp->s_msgxmitsz > net.n_maxsdu) {
		pbp = dupmp->b_cont;
		bp = pbp->b_cont;
		size = sizeof(hdr_t);
		while (size < (int)stp->s_totxmitsz) {
			skip = stp->s_totxmitsz - size;
			if ((bp->b_wptr - bp->b_rptr) <= skip) {
				size += bp->b_wptr - bp->b_rptr;
				pbp->b_cont = bp->b_cont;
				freeb(bp);
				bp = pbp->b_cont;
			} else {
				size += skip;
				bp->b_rptr += skip;
			}
		}
		size = sizeof(hdr_t);
		while ((size < net.n_maxsdu) && (bp != NULL)) {
			maxsz = net.n_maxsdu - size;
			if ((bp->b_wptr - bp->b_rptr) <= maxsz) {
				size += bp->b_wptr - bp->b_rptr;
				bp = bp->b_cont;
			} else {
				bp->b_wptr = bp->b_rptr + maxsz;
				size += maxsz;
			}
		}
		if ((bp != NULL) && (bp->b_cont != NULL)) {
			freemsg(bp->b_cont);
			bp->b_cont = NULL;
		}
		size = msgdsize(dupmp);
		hp->h_size = htons(size);
		if (size < net.n_minsdu) {
			size = net.n_minsdu - size;
			bp = allocb(size, BPRI_HI);
			if (bp == NULL) {
				freemsg(dupmp);
				return;
			}
			bp->b_wptr += size;
			linkb(dupmp, bp);
		}
	}
	if (!retransmit) {
		s = splstr();
		stp->s_totxmitsz += ntohs(hp->h_size);
		if (mp == NULL)
			stp->s_totxmitsz -= sizeof(hdr_t);
		splx(s);
		hp->h_seqno = htonl(stp->s_seqno++);
		if (stp->s_totxmitsz == stp->s_msgxmitsz)
			hp->h_flags |= EOM;
	} else {
		stp->s_nretr++;
	}
	drv_getparm(LBOLT, &stp->s_txtime);
	putnext(net.n_wrq, dupmp);
}

STATIC int
stpursrv(queue_t *q)
{
	mblk_t *mp;
	stp_t *stp;
	queue_t *rq;
	hdr_t *hp;
	int wasflowctled, s;

	stp = (stp_t *)q->q_ptr;
	s = splstr();
	wasflowctled = stp->s_flag & FLOWCTL;
	stp->s_flag &= ~FLOWCTL;
	splx(s);
	while ((mp = getq(q)) != NULL) {
		if (!canput(q->q_next)) {
			putbq(q, mp);
			s = splstr();
			stp->s_flag |= FLOWCTL;
			splx(s);
			break;
		}
		hp = (hdr_t *)mp->b_cont->b_rptr;
		switch (hp->h_type) {
		case DATA:
			if (stp_rcvdata(q, mp) == RETRY) {
				putbq(q, mp);
				goto checkfc;
			}
			break;

		case CONNECT:
			if (stp_rcvconnect(q, mp) == RETRY) {
				putbq(q, mp);
				goto checkfc;
			}
			break;

		case ACCEPT:
			if (stp_rcvaccept(q, mp) == RETRY) {
				putbq(q, mp);
				goto checkfc;
			}
			break;

		case DISCON:
			if (stp_rcvdiscon(q, mp) == RETRY) {
				putbq(q, mp);
				goto checkfc;
			}
			break;

		case RESTART:
			freemsg(mp);
			s = splstr();
			stp->s_flag &= ~NOXMIT;
			splx(s);
			qenable(WR(q));
			break;

		default:
			freemsg(mp);
		}
	}

checkfc:
	if (wasflowctled && !(stp->s_flag & FLOWCTL)) {
		if (net.n_state == UP) {
			rq = RD(net.n_wrq);
			if (rq->q_first != NULL)
				qenable(rq);
		}
		if (stp_restart(stp) == RETRY) {
			s = splstr();
			stp->s_flag |= DORESTART;
			splx(s);
			timeout((void (*)())stp_restart, (caddr_t)stp,
			  drv_usectohz(HALFSEC));
		}
	}
	return(0);
}

STATIC void
stp_rsched(stp_t *stp, int size)
{
	ulong_t id;
	int s;

	if (stp->s_flag & (RRECBCALL|RRECTOUT))
		return;
	id = bufcall(size, BPRI_HI, stp_rcont, stp);
	s = splstr();
	if (id == 0) {
		id = timeout(stp_rcont, (caddr_t)stp,
		  drv_usectohz(HALFSEC));
		stp->s_flag |= RRECTOUT;
	} else {
		stp->s_flag |= RRECBCALL;
	}
	splx(s);
	stp->s_rrecid = id;
}

STATIC void
stp_rcont(stp_t *stp)
{
	if (stp->s_rdq != NULL) {
		stp->s_flag &= ~(RRECBCALL|RRECTOUT);
		qenable(stp->s_rdq);
	}
}

STATIC int
stplrput(queue_t *q, mblk_t *mp)
{
	union DL_primitives *dp;
	hdr_t *hp;
	stp_t *stp;
	int didack;
	ulong_t seqno;

	switch (mp->b_datap->db_type) {
	case M_PROTO:
	case M_PCPROTO:
		dp = (union DL_primitives *)mp->b_rptr;
		switch (dp->dl_primitive) {
		case DL_INFO_ACK:
			stp_rcvinfo(mp);
			break;

		case DL_UNITDATA_IND:
			if (net.n_state != UP)
				goto drop;
			if (!pullupmsg(mp->b_cont, -1))
				goto drop; /* source will retransmit */
			hp = (hdr_t *)mp->b_cont->b_rptr;
			stp = stp_findport(ntohs(hp->h_dport));
			if (stp == NULL)
				goto drop;
			seqno = ntohl(hp->h_seqno);
			if (hp->h_type == ACK) {
				if (seqno == (stp->s_seqno - 1)) {
					stp->s_rcvack = seqno;
					if (stp->s_totxmitsz ==
					  stp->s_msgxmitsz) {
						freemsg(stp->s_rxmp);
						stp->s_rxmp = NULL;
					}
					if (hp->h_flags & STOP)
						stp->s_flag |= NOXMIT;
					else
						qenable(WR(stp->s_rdq));
				}
				goto drop;
			}
			if (stp->s_tpstate == TS_DATA_XFER) {
				if (seqno == stp->s_ackno) {	/* lost ack */
					if (stp->s_flag & FLOWCTL)
						stp_sendack(stp, mp, 1);
					else
						stp_sendack(stp, mp, 0);
					goto drop;
				} else if (seqno != (stp->s_ackno + 1)) {
					goto drop;
				}
			}
			if (!canput(stp->s_rdq->q_next)) {
				stp->s_flag |= FLOWCTL;
				didack = stp_sendack(stp, mp, 1);
			} else {
				didack = stp_sendack(stp, mp, 0);
			}
			if (!didack)
				freemsg(mp);
			else
				putq(q, mp);
			break;

		default:
drop:
			freemsg(mp);
		}
		break;

	case M_FLUSH:	/* TODO: mention about s_rcvmp */
		if (*mp->b_rptr & FLUSHR) {
			flushq(q, FLUSHALL);
			*mp->b_rptr &= ~FLUSHR;
		}
		if (*mp->b_rptr & FLUSHW)
			qreply(q, mp);
		else
			freemsg(mp);
		break;

	case M_ERROR:
		freemsg(mp);
		net.n_state = DOWN;
		stp_disconall(DIS_NETERR);
		putctl1(WR(q)->q_next, M_FLUSH, FLUSHRW);
		break;

	default:
		freemsg(mp);
	}
	return(0);
}

STATIC int
stp_sendack(stp_t *stp, mblk_t *mp, uchar_t stop)
{
	hdr_t *hp;
	mblk_t *hbp, *dbp;
	dl_unitdata_ind_t *indp;
	struct stpaddr addr;
	int size;
	ulong_t seqno;
	uchar_t flags = EOM;

	hp = (hdr_t *)mp->b_cont->b_rptr;
	seqno = hp->h_seqno;
	indp = (dl_unitdata_ind_t *)mp->b_rptr;
	stp_buildaddr(&addr, (caddr_t)mp->b_rptr+
	  indp->dl_src_addr_offset, hp->h_sport);
	size = HDRSIZE;
	hbp = allocb(size, BPRI_HI);
	if (hbp == NULL)
		return(0);
	dbp = allocb(DL_UNITDATA_REQ_SIZE+net.n_addrlen,
	  BPRI_HI);
	if (dbp == NULL) {
		freeb(hbp);
		return(0);
	}
	stp_buildunitdata(dbp, &addr);
	if (stop)
		flags |= STOP;
	stp_buildheader(stp, hbp, ACK, flags, 0, sizeof(hdr_t),
	  addr.a_port, size);
	hp = (hdr_t *)hbp->b_rptr;
	hp->h_seqno = seqno;
	dbp->b_cont = hbp;
	stp->s_ackno = ntohl(seqno);
	putnext(net.n_wrq, dbp);
	return(1);
}

STATIC void
stp_buildaddr(struct stpaddr *ap, caddr_t nap,
	ushort_t port)
{
	caddr_t p = nap;

	ap->a_port = port;
	if (net.n_saplen > 0)	/* sap followed by phys */
		p += net.n_saplen;
	bcopy(p, (caddr_t)ap->a_net, NETADDRSZ);
}

STATIC void
stp_rcvinfo(mblk_t *mp)
{
	dl_info_ack_t *ackp;
	struct iocblk *iocp;
	struct linkblk *lp;
	mblk_t *bp;

	bp = net.n_mp;
	net.n_mp = NULL;
	iocp = (struct iocblk *)bp->b_rptr;
	lp = (struct linkblk *)bp->b_cont->b_rptr;
	if ((iocp->ioc_cmd == I_PLINK) &&
	  !stp_qinuse(lp->l_qtop)) {
		freemsg(bp);
		freemsg(mp);
		net.n_state = DOWN;
		return;
	}
	ackp = (dl_info_ack_t *)mp->b_rptr;
	if ((ackp->dl_service_mode != DL_CLDLS) ||
	  (ackp->dl_current_state != DL_IDLE)) {
		iocp->ioc_error = EPROTO;
		goto nak;
	}
	if (ackp->dl_addr_length > (NETADDRSZ+MAXNETSAPSZ)) {
		iocp->ioc_error = ENAMETOOLONG;
		goto nak;
	}
	if (ackp->dl_max_sdu < (sizeof(hdr_t)+1)) {
		iocp->ioc_error = E2BIG;
nak:
		bp->b_datap->db_type = M_IOCNAK;
		qreply(lp->l_qtop, bp);
		freemsg(mp);
		net.n_state = DOWN;
		return;
	}
	net.n_minsdu = ackp->dl_min_sdu;
	net.n_maxsdu = ackp->dl_max_sdu;
	net.n_addrlen = ackp->dl_addr_length;
	net.n_saplen = ackp->dl_sap_length;
	bcopy((caddr_t)mp->b_rptr + ackp->dl_addr_offset,
	  (caddr_t)net.n_addr, net.n_addrlen);
	iocp->ioc_count = 0;
	bp->b_datap->db_type = M_IOCACK;
	qreply(lp->l_qtop, bp);
	freemsg(mp);
	net.n_state = UP;
	timeout(stp_retransmit, 0, drv_usectohz(RETRTIME));
}

STATIC int
stp_qinuse(queue_t *q)
{
	int i;

	for (i = 0; i < nstp; i++)
		if (stpendpt[i].s_rdq == RD(q))
			return(1);
	return(0);
}

STATIC int
stp_samenetaddr(uchar_t *a1, uchar_t *a2)
{
	int i;

	for (i = 0; i < NETADDRSZ; i++)
		if (*a1++ != *a2++)
			return(0);
	return(1);
}

STATIC retval_t
stp_rcvdata(queue_t *q, mblk_t *mp)
{
	stp_t *stp;
	mblk_t *bp, *udbp;
	hdr_t *hp;
	dl_unitdata_ind_t *udp;
	struct T_data_ind *indp;
	struct stpaddr addr;
	int eom;

	stp = (stp_t *)q->q_ptr;
	if (stp->s_tpstate != TS_DATA_XFER) {
		freemsg(mp);
		return(ERR);
	}
	udp = (dl_unitdata_ind_t *)mp->b_rptr;
	stp_buildaddr(&addr, (caddr_t)mp->b_rptr+
	  udp->dl_src_addr_offset, 0);
	if (!stp_samenetaddr(stp->s_faddr.a_net, addr.a_net)) {
		freemsg(mp);
		return(ERR);
	}
	udbp = mp;
	mp = mp->b_cont;
	hp = (hdr_t *)mp->b_rptr;
	eom = hp->h_flags & EOM;
	if ((mp->b_wptr - mp->b_rptr) != ntohs(hp->h_size))
		mp->b_wptr = mp->b_rptr + ntohs(hp->h_size);
	mp->b_rptr += sizeof(hdr_t);
	if (stp->s_rcvmp == NULL) {
		bp = allocb(sizeof(struct T_data_ind), BPRI_MED);
		if (bp == NULL) {
			mp->b_rptr -= sizeof(hdr_t);
			stp_rsched(stp, sizeof(struct T_data_ind));
			return(RETRY);
		}
		bp->b_cont = mp;
		indp = (struct T_data_ind *)bp->b_wptr;
		indp->PRIM_type = T_DATA_IND;
		indp->MORE_flag = hp->h_flags & MORE;
		bp->b_wptr += sizeof(struct T_data_ind);
		bp->b_datap->db_type = M_PROTO;
		stp->s_rcvmp = bp;
	} else {
		linkb(stp->s_rcvmp, mp);
	}
	freeb(udbp);
	if (eom) {
		putnext(q, stp->s_rcvmp);
		stp->s_rcvmp = NULL;
	}
	return(DONE);
}

STATIC retval_t
stp_rcvconnect(queue_t *q, mblk_t *mp)
{
	stp_t *stp;
	mblk_t *bp, *cbp;
	hdr_t *hp;
	struct stpaddr addr;
	dl_unitdata_ind_t *udp;
	struct T_conn_ind *indp;
	int i, s;

	stp = (stp_t *)q->q_ptr;
	if (BADSTATE(stp->s_tpstate, TE_CONN_IND) ||
	  (stp->s_qlen == stp->s_maxqlen)) {
		freemsg(mp);
		return(ERR);
	}
	udp = (dl_unitdata_ind_t *)mp->b_rptr;
	hp = (hdr_t *)mp->b_cont->b_rptr;
	stp_buildaddr(&addr, (caddr_t)mp->b_rptr+
	  udp->dl_src_addr_offset, hp->h_sport);
	bp = allocb(sizeof(struct T_conn_ind)+STPADDRSZ,
	  BPRI_MED);
	if (bp == NULL) {
		stp_rsched(stp, sizeof(struct T_conn_ind)+
		  STPADDRSZ);
		return(RETRY);
	}
	cbp = dupb(bp);
	if (cbp == NULL) {
		freeb(bp);
		if (!(stp->s_flag & (RRECBCALL|RRECTOUT))) {
			stp->s_rrecid = timeout(stp_rcont,
			  (caddr_t)stp, drv_usectohz(HALFSEC));
			s = splstr();
			stp->s_flag |= RRECTOUT;
			splx(s);
		}
		return(RETRY);
	}
	freemsg(mp);
	indp = (struct T_conn_ind *)bp->b_wptr;
	indp->PRIM_type = T_CONN_IND;
	indp->SRC_length = STPADDRSZ;
	indp->SRC_offset = sizeof(struct T_conn_ind);
	indp->OPT_length = 0;
	indp->OPT_offset = 0;
	indp->SEQ_number = stp->s_conseq++;
	bp->b_wptr += sizeof(struct T_conn_ind);
	bcopy((caddr_t)&addr, (caddr_t)bp->b_wptr,
	  STPADDRSZ);
	bp->b_wptr += STPADDRSZ;
	bp->b_datap->db_type = M_PROTO;
	cbp->b_wptr += (bp->b_wptr - bp->b_rptr);
	for (i = 0; i < MAXCREQ; i++)
		if (stp->s_cmp[i] == NULL)
			break;
	stp->s_cmp[i] = cbp;
	stp->s_qlen++;
	stp->s_tpstate = NEXTSTATE1(stp->s_tpstate,
	  TE_CONN_IND);
	putnext(q, bp);
	return(DONE);
}

STATIC retval_t
stp_rcvaccept(queue_t *q, mblk_t *mp)
{
	stp_t *stp;
	mblk_t *bp;
	hdr_t *hp;
	struct stpaddr addr;
	dl_unitdata_ind_t *udp;
	struct T_conn_con *conp;

	stp = (stp_t *)q->q_ptr;
	if (BADSTATE(stp->s_tpstate, TE_CONN_CON)) {
		freemsg(mp);
		return(ERR);
	}
	udp = (dl_unitdata_ind_t *)mp->b_rptr;
	stp_buildaddr(&addr, (caddr_t)mp->b_rptr+
	  udp->dl_src_addr_offset, 0);
	if (!stp_samenetaddr(stp->s_faddr.a_net, addr.a_net)) {
		freemsg(mp);
		return(ERR);
	}
	bp = allocb(sizeof(struct T_conn_con)+STPADDRSZ,
	  BPRI_MED);
	if (bp == NULL) {
		stp_rsched(stp, sizeof(struct T_conn_con)+
		  STPADDRSZ);
		return(RETRY);
	}
	hp = (hdr_t *)mp->b_cont->b_rptr;
	conp = (struct T_conn_con *)bp->b_wptr;
	conp->PRIM_type = T_CONN_CON;
	conp->RES_length = STPADDRSZ;
	conp->RES_offset = sizeof(struct T_conn_con);
	conp->OPT_length = 0;
	conp->OPT_offset = 0;
	bp->b_wptr += sizeof(struct T_conn_con);
	addr.a_port = hp->h_sport;
	freemsg(mp);
	bcopy((caddr_t)&addr, (caddr_t)bp->b_wptr,
	  STPADDRSZ);
	stp->s_faddr.a_port = addr.a_port;
	bp->b_wptr += STPADDRSZ;
	bp->b_datap->db_type = M_PROTO;
	stp->s_tpstate = NEXTSTATE1(stp->s_tpstate,
	  TE_CONN_CON);
	putnext(q, bp);
	return(DONE);
}

STATIC retval_t
stp_rcvdiscon(queue_t *q, mblk_t *mp)
{
	stp_t *stp;
	mblk_t *bp;
	hdr_t *hp;
	struct stpaddr addr;
	dl_unitdata_ind_t *udp;
	struct T_conn_ind *cindp;
	struct T_discon_ind *indp;
	int i;

	stp = (stp_t *)q->q_ptr;
	bp = allocb(sizeof(struct T_discon_ind), BPRI_MED);
	if (bp == NULL) {
		stp_rsched(stp, sizeof(struct T_discon_ind));
		return(RETRY);
	}
	indp = (struct T_discon_ind *)bp->b_wptr;
	indp->PRIM_type = T_DISCON_IND;
	bp->b_wptr += sizeof(struct T_discon_ind);
	bp->b_datap->db_type = M_PROTO;
	hp = (hdr_t *)mp->b_cont->b_rptr;
	udp = (dl_unitdata_ind_t *)mp->b_rptr;
	stp_buildaddr(&addr, (caddr_t)mp->b_rptr +
	  udp->dl_src_addr_offset, hp->h_sport);
	indp->DISCON_reason = hp->h_reason;
	freemsg(mp);
	if (stp->s_tpstate == TS_WRES_CIND) {
		for (i = 0; i < MAXCREQ; i++) {
			if (stp->s_cmp[i] == NULL)
				continue;
			cindp = (struct T_conn_ind *)
			  stp->s_cmp[i]->b_rptr;
			if (stp_sameaddr(&addr,
			  (struct stpaddr *)(stp->s_cmp[i]->b_rptr +
			  cindp->SRC_offset))) {
				indp->SEQ_number = cindp->SEQ_number;
				freeb(stp->s_cmp[i]);
				stp->s_cmp[i] = NULL;
				stp->s_qlen--;
				break;
			}
		}
		if (i >= MAXCREQ) {
			freeb(bp);
			return(ERR);
		}
		if (stp->s_qlen == 0)
			stp->s_tpstate = NEXTSTATE1(stp->s_tpstate,
			  TE_DISCON_IND2);
		else
			stp->s_tpstate = NEXTSTATE1(stp->s_tpstate,
			  TE_DISCON_IND3);
	} else {
		if (!stp_samenetaddr(stp->s_faddr.a_net,
		  addr.a_net)) {
			freeb(bp);
			return(ERR);
		}
		if (BADSTATE(stp->s_tpstate, TE_DISCON_IND1)) {
			freeb(bp);
			return(ERR);
		}
		indp->SEQ_number = -1;
		if (stp->s_tpstate == TS_DATA_XFER)
			putctl1(q->q_next, M_FLUSH, FLUSHRW);
		stp->s_tpstate = NEXTSTATE1(stp->s_tpstate,
		  TE_DISCON_IND1);
	}
	putnext(q, bp);
	return(DONE);
}

STATIC int
stplrsrv(queue_t *q)
{
	mblk_t *mp;
	stp_t *stp;
	hdr_t *hp;

	if (net.n_state != UP)	{
		flushq(q, FLUSHALL);
		return(0);
	}
	while ((mp = getq(q)) != NULL) {
		hp = (hdr_t *)mp->b_cont->b_rptr;
		stp = stp_findport(ntohs(hp->h_dport));
		if (stp == NULL) {
			freemsg(mp);
		} else if (canput(stp->s_rdq)) {
			putq(stp->s_rdq, mp);
		} else {
			putbq(q, mp);
			break;
		}
	}
	return(0);
}

STATIC int
stplwsrv(queue_t *q)
{
	int i;
	queue_t *wq;

	for (i = net.n_nextsend; i < nstp; i++) {
		if (stpendpt[i].s_rdq != NULL) {
			wq = WR(stpendpt[i].s_rdq);
			if (wq->q_first != NULL)
				qenable(wq);
		}
	}
	for (i = 0; i < net.n_nextsend; i++) {
		if (stpendpt[i].s_rdq != NULL) {
			wq = WR(stpendpt[i].s_rdq);
			if (wq->q_first != NULL)
				qenable(wq);
		}
	}
	if ((++net.n_nextsend) == nstp)
		net.n_nextsend = 0;
	return(0);
}

STATIC stp_t *
stp_findport(ushort_t port)
{
	int i;
	stp_t *stp;

	stp = stpendpt;
	for (i = 0; i < nstp; i++, stp++)
		if ((stp->s_rdq != NULL) &&
		  (stp->s_lport == port))
			return(stp);
	return(NULL);
}

STATIC ushort_t
stp_getuport(void)
{
	int i;

	for (i = 1; i < MAXPORT; i++)
		if (stp_findport(i) == NULL)
			return(i);
	return(0);
}

STATIC void
stp_retransmit(void)
{
	int i;
	ulong_t now;

	if (net.n_state != UP)
		return;
	for (i = net.n_nextsend; i < nstp; i++) {
		if (stpendpt[i].s_rxmp != NULL) {
			drv_getparm(LBOLT, &now);
			if ((now - stpendpt[i].s_txtime) >=
			  drv_usectohz(RETRTIME)) {
				if (stpendpt[i].s_nretr == MAXRETR) {
					stp_discon1(&stpendpt[i],
					  DIS_TIMEOUT);
				} else {
					stp_xmit(&stpendpt[i], NULL, 1);
				}
			}
		}
	}
	for (i = 0; i < net.n_nextsend; i++) {
		if (stpendpt[i].s_rxmp != NULL) {
			drv_getparm(LBOLT, &now);
			if ((now - stpendpt[i].s_txtime) >=
			  drv_usectohz(RETRTIME)) {
				if (stpendpt[i].s_nretr == MAXRETR) {
					stp_discon1(&stpendpt[i],
					  DIS_TIMEOUT);
				} else {
					stp_xmit(&stpendpt[i], NULL, 1);
				}
			}
		}
	}
	if (++net.n_nextsend == nstp)
		net.n_nextsend = 0;
	timeout(stp_retransmit, 0, drv_usectohz(RETRTIME));
}

STATIC void
stp_discon1(stp_t *stp, uchar_t reason)
{
	mblk_t *bp;
	struct T_discon_ind *indp;

	if (stp->s_rdq == NULL)
		return;
	if ((stp->s_tpstate == TS_DATA_XFER) ||
	  (stp->s_tpstate == TS_WCON_CREQ)) {
		if ((stp->s_tpstate == TS_DATA_XFER) && putctl1(
		  stp->s_rdq->q_next, M_FLUSH, FLUSHRW) == 0)
			return;
		bp = allocb(sizeof(struct T_discon_ind), BPRI_HI);
		if (bp == NULL)
			return;
		stp->s_tpstate = NEXTSTATE1(stp->s_tpstate,
		  TE_DISCON_IND1);
		bp->b_datap->db_type = M_PROTO;
		bp->b_wptr += sizeof(struct T_discon_ind);
		indp = (struct T_discon_ind *)bp->b_rptr;
		indp->PRIM_type = T_DISCON_IND;
		indp->DISCON_reason = reason;
		indp->SEQ_number = -1;
		putnext(stp->s_rdq, bp);
	}
	if (stp->s_rxmp != NULL) {
		freemsg(stp->s_rxmp);
		stp->s_rxmp = NULL;
	}
	if (stp->s_rcvmp != NULL) {
		freemsg(stp->s_rcvmp);
		stp->s_rcvmp = NULL;
	}
}

STATIC void
stp_disconall(uchar_t reason)
{
	mblk_t *bp, *mp;
	stp_t *stp;
	struct T_discon_ind *indp;
	int i;

	bp = allocb(sizeof(struct T_discon_ind), BPRI_HI);
	if (bp == NULL)
		return;
	bp->b_datap->db_type = M_PROTO;
	bp->b_wptr += sizeof(struct T_discon_ind);
	indp = (struct T_discon_ind *)bp->b_rptr;
	indp->PRIM_type = T_DISCON_IND;
	indp->DISCON_reason = reason;
	indp->SEQ_number = -1;
	for (i = 0, stp = stpendpt; i < nstp; i++, stp++) {
		if ((stp->s_rdq != NULL) && ((stp->s_tpstate ==
		  TS_DATA_XFER) || (stp->s_tpstate ==
		  TS_WCON_CREQ))) {
			if ((stp->s_tpstate == TS_DATA_XFER) &&
			  putctl1(stp->s_rdq->q_next, M_FLUSH,
			  FLUSHRW) == 0)
				continue;
			mp = dupb(bp);
			if (mp == NULL)
				continue;
			stp->s_tpstate = NEXTSTATE1(stp->s_tpstate,
			  TE_DISCON_IND1);
			putnext(stp->s_rdq, mp);
			if (stp->s_rxmp != NULL) {
				freemsg(stp->s_rxmp);
				stp->s_rxmp = NULL;
			}
			if (stp->s_rcvmp != NULL) {
				freemsg(stp->s_rcvmp);
				stp->s_rcvmp = NULL;
			}
		}
	}
	freeb(bp);
}
