/*
 * Copyright (c) 1993 Addison-Wesley Publishing Company.
 * All rights reserved.  This source code is provided
 * for its instructional value, and is not guaranteed
 * for any particular purpose.
 */
#include <sys/types.h>
#include <signal.h>
#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <wait.h>

#define INPUTSIZ	256

static sigjmp_buf env;

static void catcher(int);
static void nop(int);
void doshell(char *);
extern void error(const char *fmt, ...);
extern void fatal(const char *fmt, ...);

void
main()
{
	struct sigaction sa;
	char buf[INPUTSIZ];

	/*
	 * Protect against a siglongjmp before the real
	 * sigsetjmp is done.
	 */
	if (sigsetjmp(env, 0)) {
		fprintf(stderr, "Interrupted\n");
		exit(1);
	}

	/*
	 * Print the initial prompt.
	 */
	printf("> ");

	/*
	 * Set up a signal handler for SIGINT and SIGQUIT.
	 */
	sa.sa_handler = catcher;
	sigemptyset(&sa.sa_mask);
	sigaddset(&sa.sa_mask, SIGINT);
	sigaddset(&sa.sa_mask, SIGQUIT);
	sa.sa_flags = 0;
	if (sigaction(SIGINT, &sa, NULL) < 0)
		fatal("sigaction failed");
	if (sigaction(SIGQUIT, &sa, NULL) < 0)
		fatal("sigaction failed");

	/*
	 * This is the real sigsetjmp.  If it returns a
	 * nonzero value, then the signal handler called
	 * siglongjmp.  Redisplay the prompt in this case.
	 */
	if (sigsetjmp(env, 1))
		printf("\n> ");

	/*
	 * While the user is providing input, read a command,
	 * execute it in a subshell, and print the next prompt.
	 */
	while (fgets(buf, INPUTSIZ, stdin) != NULL) {
		doshell(buf);
		printf("> ");
	}
	exit(0);
}

/*
 * Signal handler for SIGINT and SIGQUIT.
 */
static void
catcher(int sig)
{
	/*
	 * Flush the input and output, and jump back to
	 * print the prompt again.
	 */
	fflush(stdout);
	fflush(stdin);
	siglongjmp(env, 1);
}

void
doshell(char *cmd)
{
	pid_t pid, wpid;
	struct sigaction sigi, osigi;
	struct sigaction sigq, osigq;

	/*
	 * Install a dummy signal handler so we don't jump
	 * back to print the prompt until we are done.
	 */
	sigemptyset(&sigi.sa_mask);
	sigi.sa_handler = nop;
	sigi.sa_flags = 0;
	if (sigaction(SIGINT, &sigi, &osigi) < 0)
		fatal("sigaction failed");
	sigemptyset(&sigq.sa_mask);
	sigq.sa_handler = nop;
	sigq.sa_flags = 0;
	if (sigaction(SIGQUIT, &sigq, &osigq) < 0)
		fatal("sigaction failed");

	/*
	 * Create a child process to execute the command.
	 */
	if ((pid = fork()) < 0) {
		error("fork failed");

		/*
		 * Restore the prior signal dispositions.
		 */
		if (sigaction(SIGINT, &osigi, NULL) < 0)
			fatal("sigaction failed");
		if (sigaction(SIGQUIT, &osigq, NULL) < 0)
			fatal("sigaction failed");
		return;
	}
	if (pid == 0) {	/* child */
		/*
		 * The child executes the command in a
		 * subshell.  This implies the command
		 * can be a shell script.
		 */
		execl("/sbin/sh", "sh", "-c", cmd, NULL);
		fatal("exec failed");
	}

	/*
	 * The parent waits for the child to exit before
	 * restoring the prior signal dispositions and
	 * returning.
	 */
	while ((wpid = waitpid(pid, NULL, 0)) != pid)
		if ((wpid == -1) && (errno != EINTR))
			fatal("waitpid failed");
	if (sigaction(SIGINT, &osigi, NULL) < 0)
		fatal("sigaction failed");
	if (sigaction(SIGQUIT, &osigq, NULL) < 0)
		fatal("sigaction failed");
}

/*
 * This is a dummy signal handler.
 */
static void
nop(int sig)
{
}
