/*
 * Copyright (c) 1993 Addison-Wesley Publishing Company.
 * All rights reserved.  This source code is provided
 * for its instructional value, and is not guaranteed
 * for any particular purpose.
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include <netinet/in.h>
#include <netdb.h>
#include <utime.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#define NEWSTIME	"/var/.newstime"
#define NEWSPERM	(S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)
#define NETBUFSZ	8192

struct newshost {
	struct newshost		*nh_next;
	char				*nh_name;
	struct in_addr		nh_addr;
};

char *service = "newsd";
struct newshost *hostlist;
char lastrecv[MAXNAMELEN];
ushort_t newsport;
time_t scantime;
struct utsname u;
void gethosts(void);
void broadcast(char *, struct stat *);
void recvnews(int);
void newstime(int);
void doscan(void);

extern void daemonize(const char *);
extern void log(const char *, ...);

void
main(void)
{
	struct stat sbuf;
	struct sockaddr_in addr;
	int fd, nfd, i;
	struct timeval t;
	fd_set rset;
	struct servent *sp;

	/*
	 * Become a daemon.
	 */
	daemonize(service);

	/*
	 * Change the current working directory to
	 * the location where news items are stored.
	 */
	if (chdir("/var/news") < 0) {
		log("cannot cd to /var/news: %s",
		  strerror(errno));
		exit(1);
	}

	/*
	 * Get the local host name.
	 */
	if (uname(&u) < 0) {
		log("uname failed: %s", strerror(errno));
		exit(1);
	}

	/*
	 * Check the last time /var/news was scanned.
	 */
	if (stat(NEWSTIME, &sbuf) < 0) {
		if (errno == ENOENT) {
			/*
			 * /var/.newstime doesn't exist --
			 * create it.
			 */
			if ((fd = creat(NEWSTIME, NEWSPERM)) < 0) {
				log("cannot create %s: %s", NEWSTIME,
				  strerror(errno));
				exit(1);
			}
			close(fd);
		} else {
			/*
			 * stat failed for an unexpected reason.
			 */
			log("cannot stat %s: %s", NEWSTIME,
			  strerror(errno));
			exit(1);
		}
	}
	/*
	 * Create a list of hosts and addresses
	 * from /etc/hosts.
	 */
	gethosts();

	/*
	 * Create a socket to use when sending and
	 * receiving news files.
	 */
	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		log("cannot create socket: %s", strerror(errno));
		exit(1);
	}

	/*
	 * Identify the port number to use for the
	 * news service.
	 */
	sp = getservbyname(service, NULL);
	if (!sp) {
		log("cannot find service for %s", service);
		exit(1);
	}

	/*
	 * Create the address and bind it to the socket.
	 */
	newsport = sp->s_port;
	addr.sin_family = AF_INET;
	addr.sin_port = newsport;
	addr.sin_addr.s_addr = INADDR_ANY;
	for (i = 0; i < sizeof(addr.sin_zero); i++)
		addr.sin_zero[i] = 0;
	if (bind(fd, &addr, sizeof(struct sockaddr_in)) < 0) {
		log("cannot bind address: %s", strerror(errno));
		exit(1);
	}

	/*
	 * Enable the socket to be used to receive
	 * connect indications.
	 */
	if (listen(fd, 10) < 0) {
		log("cannot listen: %s", strerror(errno));
		exit(1);
	}

	/*
	 * Read the time of the last scan of /var/news.
	 */
	newstime(0);

	/*
	 * Scan /var/news for news items to transmit.
	 */
	doscan();
	/*
	 * This is the main processing loop.  Wait
	 * for connect indications for 30 seconds.
	 * If none arrive, scan /var/news for files
	 * to transmit.
	 */
	for (;;) {
		/*
		 * Use select to avoid blocking indefinitely.
		 */
		t.tv_sec = 30;
		t.tv_usec = 0;
		FD_ZERO(&rset);
		FD_SET(fd, &rset);
		i = select(fd+1, &rset, NULL, NULL, &t);
		if (i < 0) {
			log("select failed: %s", strerror(errno));
			exit(1);
		} else if (i == 0) {
			/*
			 * No network traffic -- scan /var/news.
			 */
			doscan();
			continue;
		} else if ((i != 1) || !FD_ISSET(fd, &rset)) {
			/*
			 * This shouldn't happen.
			 */
			log("select ret %d, isset %d; unknown error",
			  i, FD_ISSET(fd, &rset));
			exit(1);
		}

		/*
		 * We know fd is now readable, meaning that
		 * a connect indication has arrived.  Accept it.
		 */
		nfd = accept(fd, NULL, NULL);
		if (nfd < 0) {
			log("cannot accept connection: %s",
			  strerror(errno));
			exit(1);
		}

		/*
		 * Receive the news file and store it in
		 * /var/news.  Then close the newly created
		 * network connection.
		 */
		recvnews(nfd);
		close(nfd);

		/*
		 * Before waiting for more connect indications,
		 * check /var/news for newly created news items.
		 */
		doscan();
	}
} /* end of main */

void
gethosts(void)
{
	struct hostent *hp;
	struct newshost *nhp;
	struct in_addr *ip;

	/*
	 * Read /etc/hosts one line at a time.
	 */
	while ((hp = gethostent()) != NULL) {
		/*
		 * Skip any entries that represent the
		 * local host.
		 */
		if ((strcmp(hp->h_name, "localhost") == 0) ||
		  (strcmp(hp->h_name, "anyhost") == 0) ||
		  (strcmp(hp->h_name, u.nodename) == 0))
			continue;

		/*
		 * Allocate a newshost structure to represent
		 * this entry.
		 */
		nhp = malloc(sizeof(struct newshost));
		if (nhp == NULL) {
			log("cannot allocate memory for host list");
			exit(1);
		}

		/*
		 * Initialize the newshost structure with the
		 * address and name of the host.  Then put the
		 * structure on the front of the list of hosts.
		 */
		ip = (struct in_addr *)hp->h_addr;
		nhp->nh_addr.s_addr = ip->s_addr;
		nhp->nh_name = strdup(hp->h_name);
		nhp->nh_next = hostlist;
		hostlist = nhp;
	}

	/*
	 * Close /etc/hosts and clean up.
	 */
	endhostent();
}

void
doscan(void)
{
	DIR *dp;
	struct dirent *dep;
	struct stat sbuf;
	time_t lastscan;

	/*
	 * Remember the last time we scanned /var/news.
	 */
	lastscan = scantime;

	/*
	 * Update /var/news with the current time.
	 */
	newstime(1);

	/*
	 * Open the current directory (/var/news).
	 */
	dp = opendir(".");
	if (dp == NULL) {
		log("cannot open directory: %s",
		  strerror(errno));
		exit(1);
	}

	/*
	 * Use readdir to read one directory entry at a time.
	 */
	while ((dep = readdir(dp)) != NULL) {
		/*
		 * Skip the directory entries for dot and
		 * dot-dot.  Also skip the entry for the
		 * last file received from the network.
		 */
		if ((strcmp(dep->d_name, ".") == 0) ||
		  (strcmp(dep->d_name, "..") == 0) ||
		  (strcmp(dep->d_name, lastrecv) == 0))
			continue;

		/*
		 * Get the file attributes.  If this fails,
		 * just skip the file.
		 */
		if (stat(dep->d_name, &sbuf) < 0)
			continue;

		/*
		 * If the file has been modified since the last
		 * scan of /var/news, and if it is a regular
		 * file, transmit it to the other hosts.
		 */
		if ((sbuf.st_mtime > lastscan) &&
		  S_ISREG(sbuf.st_mode))
			broadcast(dep->d_name, &sbuf);
	}

	/*
	 * Close /var/news.
	 */
	closedir(dp);
}

void
newstime(int touchit)
{
	struct stat sbuf;

	/*
	 * If touchit is nonzero, update the modification
	 * time on /var/.newstime.
	 */
	if (touchit && (utime(NEWSTIME, NULL) < 0)) {
		log("cannot change mod time on %s: %s", NEWSTIME,
		  strerror(errno));
		exit(1);
	}

	/*
	 * Get the modification time of /var/.newstime and
	 * save it in scantime.
	 */
	if (stat(NEWSTIME, &sbuf) < 0) {
		log("cannot stat %s: %s", NEWSTIME,
		  strerror(errno));
		exit(1);
	}
	scantime = sbuf.st_mtime;
}

void
broadcast(char *filename, struct stat *stp)
{
	int fd, nfd, len, ret;
	struct newshost *nhp;
	struct sockaddr_in dest;
	char buf[NETBUFSZ];

	/*
	 * Open the file to be transmitted.
	 */
	if ((fd = open(filename, O_RDONLY)) < 0) {
		log("cannot open %s for broadcast: %s", filename,
		  strerror(errno));
		return;
	}

	/*
	 * Create most of the destination address.
	 */
	dest.sin_family = AF_INET;
	dest.sin_port = newsport;
	for (len = 0; len < sizeof(dest.sin_zero); len++)
		dest.sin_zero[len] = 0;

	/*
	 * Step through the list of hosts and transmit the
	 * file to each one.
	 */
	for (nhp = hostlist; nhp; nhp = nhp->nh_next) {

		/*
		 * Seek to the beginning of the file.
		 */
		lseek(fd, 0, SEEK_SET);

		/*
		 * Finish initializing the destination
		 * address.
		 */
		dest.sin_addr.s_addr = nhp->nh_addr.s_addr;

		/*
		 * Create a socket.
		 */
		if ((nfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
			log("cannot create socket: %s",
			  strerror(errno));
			continue;
		}

		/*
		 * Connect to the host.
		 */
		if (connect(nfd, (struct sockaddr *)&dest,
		  sizeof(struct sockaddr_in)) < 0) {
			log("cannot connect to %s: %s", nhp->nh_name,
			  strerror(errno));
			close(nfd);
			continue;
		}
		/*
		 * Calculate the length of the filename and
		 * send it to the remote host.
		 */
		len = strlen(filename)+1;
		ret = send(nfd, filename, len, 0);
		if (len != ret) {
			if (ret < 0)
				log("send failed: %s", strerror(errno));
			else
				log("sent less than requested (%d/%d)",
				  ret, len);
			close(nfd);
			continue;
		}

		/*
		 * Now read the file in NETBUFSZ chunks
		 * and transmit each chunk to the host.
		 */
		while (stp->st_size > 0) {
			len = read(fd, buf, NETBUFSZ);
			if (len < 0) {
				log("cannot read file: %s",
				  strerror(errno));
				break;
			}
			ret = send(nfd, buf, len, 0);
			if (len != ret) {
				if (ret < 0)
					log("send failed: %s",
					  strerror(errno));
				else
					log("sent less than requested (%d/%d)",
					  ret, len);
				break;
			}
			stp->st_size -= len;
		}

		/*
		 * Close the connection.
		 */
		close(nfd);
	}

	/*
	 * Close the news file.
	 */
	close(fd);
} /* end of broadcast */

void
recvnews(int nfd)
{
	int fd, n, ret;
	char *p;
	char buf[NETBUFSZ];

	/* 
	 * Read the name of the file to be received.
	 */
	p = lastrecv;
	for (;;) {
		/*
		 * Don't consume the data; just peek at it.
		 */
		ret = recv(nfd, buf, NETBUFSZ, MSG_PEEK);
		if (ret < 0) {
			log("recv failed: %s", strerror(errno));
			return;
		}

		/*
		 * Copy the filename to lastrecv.
		 */
		for (n = 0; n < ret; n++)
			if (buf[n] == '\0')
				break;
		n++;
		strncpy(p, buf, n);
		p += n;

		/*
		 * Now consume the filename.
		 */
		if ((ret = recv(nfd, buf, n, 0)) != n) {
			if (ret < 0)
				log("recv failed: %s", strerror(errno));
			else
				log("recv expected %d, returned %d", n,
				  ret);
			lastrecv[0] = '\0';
			return;
		}

		/*
		 * If we've reached the end of the filename,
		 * break out of the loop and receive the file.
		 */
		if (*(p - 1) == '\0')
			break;
	}

	/*
	 * Create the file, making it publicly readable.
	 */
	if ((fd = creat(lastrecv, NEWSPERM)) < 0) {
		log("cannot create /var/news/%s: %s", lastrecv,
		  strerror(errno));
		lastrecv[0] = '\0';
		return;
	}

	/*
	 * Read the contents of the file from the network
	 * and save it on disk.
	 */
	while ((n = recv(nfd, buf, NETBUFSZ, 0)) > 0)
		write(fd, buf, n);

	/*
	 * Close the file.
	 */
	close(fd);
}
